@@ -13,6 +13,10 @@
 #include <kapplication.h>
 #include <kdebug.h>
 
+// Default button layout
+const char default_left[]  = "X";
+const char default_right[] = "HSIA";
+
 namespace System {
 
 static unsigned char iconify_bits[] = {
@@ -21,9 +25,12 @@ static unsigned char iconify_bits[] = {
 static unsigned char maximize_bits[] = {
     0x00, 0x18, 0x3c, 0x7e, 0xff, 0xff, 0x00, 0x00};
 
-static unsigned char minmax_bits[] = {
+static unsigned char r_minmax_bits[] = {
     0x0c, 0x18, 0x33, 0x67, 0xcf, 0x9f, 0x3f, 0x3f};
 
+static unsigned char l_minmax_bits[] = {
+    0x30, 0x18, 0xcc, 0xe6, 0xf3, 0xf9, 0xfc, 0xfc};    
+    
 static unsigned char unsticky_bits[] = {
     0x00, 0x18, 0x18, 0x7e, 0x7e, 0x18, 0x18, 0x00};
 
@@ -281,13 +288,18 @@ SystemClient::SystemClient(KDecorationBridge* bridge, KDecorationFactory* factor
     : KDecoration(bridge, factory)
 {}
 
+SystemClient::~SystemClient()
+{
+    for (int n=0; n<ButtonTypeCount; n++) {
+        if (button[n]) delete button[n];
+    }
+}
+
 void SystemClient::init()
 {
    createMainWidget(0);
    widget()->installEventFilter( this );
 
-   bool help = providesContextHelp();
- 
    QGridLayout* g = new QGridLayout(widget(), 0, 0, 2);
 
    if (isPreview())
@@ -305,73 +317,98 @@ void SystemClient::init()
    g->addColSpacing(2, 2);
    g->addRowSpacing(2, 6);
 
-   button[0] = new SystemButton(this, "close", NULL, i18n("Close"));
-   button[1] = new SystemButton(this, "sticky", NULL, i18n("On all desktops"));
-   if(isOnAllDesktops())
-      button[1]->setBitmap(unsticky_bits);
-   else
-      button[1]->setBitmap(sticky_bits);
-
-
-   button[2] = new SystemButton(this, "iconify", iconify_bits, i18n("Minimize"));
-   button[3] = new SystemButton(this, "maximize", maximize_bits, i18n("Maximize"));
-   if(help)
-   {
-      button[4] = new SystemButton(this, "help", question_bits, i18n("Help"));
-      connect( button[4], SIGNAL( clicked() ), this, ( SLOT( showContextHelp() ) ) );
-   }
-   else
-      button[4] = NULL;
-
-   connect( button[0], SIGNAL( clicked() ), this, ( SLOT( closeWindow() ) ) );
-   connect( button[1], SIGNAL( clicked() ), this, ( SLOT( toggleOnAllDesktops() ) ) );
-   connect( button[2], SIGNAL( clicked() ), this, ( SLOT( minimize() ) ) );
-   connect( button[3], SIGNAL( clicked() ), this, ( SLOT( maxButtonClicked() ) ) );
-
    QBoxLayout* hb = new QBoxLayout(0, QBoxLayout::LeftToRight, 0, 0, 0);
    hb->setResizeMode(QLayout::FreeResize);
    g->addLayout( hb, 0, 1 );
-   hb->addSpacing(3);
-   if (isCloseable())
-   {
-      hb->addWidget( button[0]);
-   }
-   else
-      button[0]->hide();
-        
+   hb->addSpacing(3);   
+   
    titlebar = new QSpacerItem(10, 14, QSizePolicy::Expanding,
-                              QSizePolicy::Minimum);
-   hb->addSpacing(3);
-   hb->addItem(titlebar);
-   hb->addSpacing(3);
-   if(help)
-   {
-      hb->addWidget( button[4]);
-      hb->addSpacing(1);
-   }
-   hb->addWidget( button[1]);
-   hb->addSpacing(1);
+                              QSizePolicy::Minimum);    
 
-   if (isMinimizable())
-   {
-      hb->addWidget(button[2]);
-      hb->addSpacing(1);
-   }
-   else
-      button[2]->hide();
-
-   if (isMaximizable())
-   {
-      hb->addWidget(button[3]);
-      hb->addSpacing(3);
-   }
-   else
-      button[3]->hide();
+    // setup titlebar buttons
+    for (int n=0; n<ButtonTypeCount; n++) button[n] = 0;
+    addButtons(hb, KDecoration::options()->customButtonPositions() ? 
+                   KDecoration::options()->titleButtonsLeft() : QString(default_left));
+    hb->addSpacing(2);    
+    hb->addItem(titlebar);
+    hb->addSpacing(3);
+    addButtons(hb, KDecoration::options()->customButtonPositions() ? 
+                   KDecoration::options()->titleButtonsRight() : QString(default_right));
+    hb->addSpacing(2);
 
    widget()->setBackgroundMode(QWidget::NoBackground);
    recalcTitleBuffer();
 }
 
+void SystemClient::addButtons(QBoxLayout *hb, const QString& s)
+{
+    unsigned char *minmax_bits;
+    int l_max = KDecoration::options()->titleButtonsLeft().find('A');
+    if (s.length() > 0) {
+        for (unsigned n=0; n < s.length(); n++) {
+            switch (s[n]) {
+              case 'X': // Close button
+                  if ((!button[ButtonClose]) && isCloseable()) {
+                      button[ButtonClose] = new SystemButton(this, "close", NULL, i18n("Close"));
+                      connect( button[ButtonClose], SIGNAL( clicked() ), this, ( SLOT( closeWindow() ) ) );
+                      hb->addWidget(button[ButtonClose]);
+                      hb->addSpacing(1);
+                  }
+                  break;
+                  
+              case 'S': // Sticky button
+                  if (!button[ButtonSticky]) {
+                     button[ButtonSticky] = new SystemButton(this, "sticky", NULL, i18n("On all desktops"));
+                     if(isOnAllDesktops())
+                         button[ButtonSticky]->setBitmap(unsticky_bits);
+                     else
+                         button[ButtonSticky]->setBitmap(sticky_bits);
+                     connect( button[ButtonSticky], SIGNAL( clicked() ), this, ( SLOT( toggleOnAllDesktops() ) ) );
+                     hb->addWidget(button[ButtonSticky]);
+                     hb->addSpacing(1);
+                  }
+                  break;
+
+              case 'I': // Minimize button
+                  if ((!button[ButtonMinimize]) && isMinimizable())  {
+                      button[ButtonMinimize] = new SystemButton(this, "iconify", iconify_bits, i18n("Minimize"));
+                      connect( button[ButtonMinimize], SIGNAL( clicked() ), this, ( SLOT( minimize() ) ) );
+                      hb->addWidget(button[ButtonMinimize]);
+                      hb->addSpacing(1);
+                  }
+                  break;
+
+              case 'A': // Maximize button
+                  if ((!button[ButtonMaximize]) && isMaximizable()) {
+                      if (maximizeMode()==MaximizeFull) {
+                          if (KDecoration::options()->customButtonPositions() && (l_max>-1))
+                              minmax_bits = l_minmax_bits;
+                          else
+                              minmax_bits = r_minmax_bits;
+                          button[ButtonMaximize] = new SystemButton(this, "maximize", minmax_bits, i18n("Restore"));
+                      }
+                      else
+                          button[ButtonMaximize] = new SystemButton(this, "maximize", maximize_bits, i18n("Maximize"));
+                      connect( button[ButtonMaximize], SIGNAL( clicked() ), this, ( SLOT( maxButtonClicked() ) ) );
+                      hb->addWidget(button[ButtonMaximize]);
+                      hb->addSpacing(1);
+                  }
+                  break;
+
+              case 'H': // Help button
+                  if ((!button[ButtonHelp]) && providesContextHelp()) {
+                      button[ButtonHelp] = new SystemButton(this, "help", question_bits, i18n("Help"));
+                      connect( button[ButtonHelp], SIGNAL( clicked() ), this, ( SLOT( showContextHelp() ) ) );
+                      hb->addWidget(button[ButtonHelp]);
+                      hb->addSpacing(1);
+                  }
+                  break;
+            
+            }
+        }
+    }
+}
+
 bool SystemClient::eventFilter( QObject* o, QEvent* e )
 {
     if( o != widget())
@@ -401,17 +438,21 @@ void SystemClient::reset(unsigned long)
    titleBuffer.resize(0, 0);
    recalcTitleBuffer();
    widget()->repaint();
-   button[0]->reset();
-   button[1]->reset();
-   button[2]->reset();
-   button[3]->reset();
-   if(button[4])
-      button[4]->reset();
+   if (button[ButtonClose])
+       button[ButtonClose]->reset();
+   if (button[ButtonSticky])
+       button[ButtonSticky]->reset();
+   if (button[ButtonMinimize])
+       button[ButtonMinimize]->reset();
+   if (button[ButtonMaximize])
+       button[ButtonMaximize]->reset();
+   if (button[ButtonHelp])
+       button[ButtonHelp]->reset();
 }
 
 void SystemClient::maxButtonClicked()
 {
-   maximize( button[3]->last_button );
+   maximize( button[ButtonMaximize]->last_button );
 }
 
 void SystemClient::resizeEvent( QResizeEvent* )
@@ -569,19 +610,31 @@ void SystemClient::mouseDoubleClickEvent( QMouseEvent * e )
 
 void SystemClient::maximizeChange()
 {
-    button[3]->setBitmap((maximizeMode()==MaximizeFull) ? minmax_bits : maximize_bits);
-    button[3]->setTipText((maximizeMode()==MaximizeFull) ? i18n("Restore") : i18n("Maximize"));
+    unsigned char *minmax_bits;
+    int l_max = KDecoration::options()->titleButtonsLeft().find('A');
+    if (KDecoration::options()->customButtonPositions() && (l_max>-1))
+        minmax_bits = l_minmax_bits;
+    else
+        minmax_bits = r_minmax_bits;
+    if (button[ButtonMaximize]) {
+        button[ButtonMaximize]->setBitmap((maximizeMode()==MaximizeFull) ? minmax_bits : maximize_bits);
+        button[ButtonMaximize]->setTipText((maximizeMode()==MaximizeFull) ? i18n("Restore") : i18n("Maximize"));
+    }
 }
 
 void SystemClient::activeChange()
 {
     widget()->repaint(false);
-    button[0]->reset();
-    button[1]->reset();
-    button[2]->reset();
-    button[3]->reset();
-    if(button[4])
-        button[4]->reset();
+    if (button[ButtonClose])
+        button[ButtonClose]->reset();
+    if (button[ButtonSticky])
+        button[ButtonSticky]->reset();
+    if (button[ButtonMinimize])
+        button[ButtonMinimize]->reset();
+    if (button[ButtonMaximize])
+        button[ButtonMaximize]->reset();
+    if (button[ButtonHelp])
+        button[ButtonHelp]->reset();
 }
 
 void SystemClient::iconChange()
@@ -592,8 +645,10 @@ void SystemClient::iconChange()
 
 void SystemClient::desktopChange()
 {
-   button[1]->setBitmap(isOnAllDesktops() ? unsticky_bits : sticky_bits);
-   button[1]->setTipText(isOnAllDesktops() ? i18n("Not on all desktops") : i18n("On all desktops"));
+   if (button[ButtonSticky]) {
+       button[ButtonSticky]->setBitmap(isOnAllDesktops() ? unsticky_bits : sticky_bits);
+       button[ButtonSticky]->setTipText(isOnAllDesktops() ? i18n("Not on all desktops") : i18n("On all desktops"));
+   }
 }
 
 /*void SystemClient::stickyChange(bool on)
@@ -635,11 +690,27 @@ KDecoration *SystemDecoFactory::createDecoration( KDecorationBridge *b )
 
 bool SystemDecoFactory::reset( unsigned long changed )
 {
-   delete_pixmaps();
-   create_pixmaps();
+   System::delete_pixmaps();
+   System::create_pixmaps();
    // Ensure changes in tooltip state get applied
    resetDecorations(changed);
-   return false;
+   return true;
+}
+
+bool SystemDecoFactory::supports( Ability ability )
+{
+    switch( ability )
+    {
+        case AbilityAnnounceButtons:
+        case AbilityButtonOnAllDesktops:
+        case AbilityButtonHelp:
+        case AbilityButtonMinimize:
+        case AbilityButtonMaximize:
+        case AbilityButtonClose:
+            return true;
+        default:
+            return false;
+    };
 }
 
 QValueList<KDecorationFactory::BorderSize> SystemDecoFactory::borderSizes() const
@@ -649,7 +720,7 @@ QValueList<KDecorationFactory::BorderSize> SystemDecoFactory::borderSizes() cons
 
 }
 
-extern "C" KDecorationFactory *create_factory()
+extern "C" KDE_EXPORT KDecorationFactory *create_factory()
 {
    return new System::SystemDecoFactory();
 }
